#!/usr/local/bin/perl
#
#   nph-proxy.cgi-- CGI Proxy 1.1: an HTTP proxy in the form of a CGI script.
#     Retrieves the resource at any URL, updating embedded URLs in HTML
#     resources to point back through this script.  By default, no user info
#     is sent to the server.  Options include text-only proxying to save
#     bandwidth, cookie support, ad-filtering, and a user-defined encoding
#     of the target URL.  Runs in Perl 4 or 5.
#
#   Copyright (C) 1996, 1998-1999 by James Marshall, james@jmarshall.com
#
#   For the latest, see http://www.jmarshall.com/tools/cgiproxy/
#
#
#   CONFIGURATION:
#     None required.  Options in the "user configuration" section include:
#       . Set $TEXT_ONLY, $SUPPORT_COOKIES, and $FILTER_ADS as desired.
#       . To change the encoding format of the URL, modify the
#           &proxy_encode() and &proxy_decode() routines.  Make sure the
#           &proxy_encode() in startproxy.cgi matches the one here.  The
#           default routines are suitable for simple PATH_INFO compliance.
#       . If you're using another HTTP proxy, set $ENV{'http_proxy'} and
#           $ENV{'no_proxy'}.
#       . To make this run in Perl 4, follow the instructions above the 
#           "use Socket" line.
#       . If filtering ads, you can customize this with a few settings.
#       . Other minor config may be possible; see the user configuration
#           section.
#       . If you're using Windows, you must comment out the line
#           "alarm(600);", because Windows doesn't seem to support alarms.
#
#     This script MUST be installed as a non-parsed header (NPH) script.
#
#
#   TO USE:
#     Use the associated startproxy.cgi to start a session, or just call
#       this script with the encoded target URL in PATH_INFO.
#
#
#   I first wrote this in 1996 as an experiment to allow anonymous browsing.
#     Parts of the code have been upgraded since then, like the addition of
#     HTML 4.0 tags.  The original seed was a program I wrote for Rich
#     Morin's article in the June 1996 issue of Unix Review, online at
#     http://www.cfcl.com/tin/P/9606.html.
#
#   Confession:  I didn't originally write this with the spec for HTTP proxies
#     in mind, and there are probably some violations of the protocol.  This
#     whole thing doesn't really match the proxy model anyway, so I hereby
#     rationalize that the spec can be widely interpreted here.  If there is
#     demand, I can make it more conformant.
#
#   TO DO:
#     more error checking
#     find a simple encryption technique for proxy_encode()
#     make able to be a real HTTP proxy, listening to a port (maybe make
#       modified version)
#     ultimately support more protocols, like SSL, FTP, or gopher
#     support streaming media (probably requires HTTP 1.1)
#     any way to support basic authentication?
#     one user wanted $TEXT_ONLY to be dependent on the host
#
#   CHANGES/FIXES/ADDITIONS IN THIS VERSION:
#     Whole format of target URL in PATH_INFO was restructured.  It can
#       be encoded however the user wishes.  This gets around PATH_INFO 
#       clashes in various servers, solving most problems regarding server 
#       incompatibilities I've heard about.
#     Cookies are optionally supported (but off by default).
#     Banner ads can be filtered (only a simple set of patterns by default)
#     Cookies from ad servers are filtered out (at least the main ones)
#     Binary files are no longer getting messed up on Windows.
#     More HTTP headers are fixed to point back through the proxy.
#     Under some conditions, extra processes would hang around for hours
#       and drag the system.  Alex Freed added a timeout to solve this
#       for now.  I can't reproduce the problem, so any info is appreciated.
#     Internally: code was cleaned up, URL-parsing was improved, and
#       relative URL calculation was redone.
#
#

#------ user configuration ------------------------------------------------

# If set, then proxy traffic will be restricted to text data only, to save 
#   bandwidth (though it can still be circumvented with uuencode, etc.).
$TEXT_ONLY= 1 ;      # set to 1 to allow only text data, 0 to allow all

# If set, then support cookies through the proxy.
$SUPPORT_COOKIES= 1 ;

# If set, then adopt several tactics to minimize ads and ad-related cookies.
# If you want to fine-tune the behavior, configure $NO_COOKIE_WITH_IMAGE, 
#   @BANNED_COOKIE_SERVERS, and @BANNED_IMAGE_URL_PATTERNS below.
$FILTER_ADS= 1 ;


# Create your own proxy_encode() and proxy_decode() to tranform the target
#   URL to and from the format that will be stored in PATH_INFO.  The encoded
#   form should only contain characters that are legal in PATH_INFO, i.e.
#   most printable chars except ? and #, and no two adjacent slashes ("//").
#   Try not to make the resulting string too long, either.
# Of course, proxy_decode() must undo whatever proxy_encode() does.
# The proxy_encode() here must match the one in startproxy.cgi.
# Make proxy_encode() as fast as possible-- it's a major bottleneck for the 
#   whole program.

sub proxy_encode {
    local($URL)= @_ ;
    $URL=~ s#^([\w+.-]+)://#$1/# ;                 # http://xxx -> http/xxx
#    $URL=~ s/(.)/ sprintf('%02x',ord($1)) /ge ;   # each char -> 2-hex
#    $URL=~ tr/a-zA-Z/n-za-mN-ZA-M/ ;              # rot-13
    return $URL ;
}

sub proxy_decode {
    local($PATH_INFO)= @_ ;
#    $PATH_INFO=~ tr/a-zA-Z/n-za-mN-ZA-M/ ;        # rot-13
#    $PATH_INFO=~ s/([0-9A-Fa-f]{2})/ sprintf("%c",hex($1)) /ge ;
    $PATH_INFO=~ s#^([\w+.-]+)/#$1://# ;           # http/xxx -> http://xxx
    return $PATH_INFO ;
}



# set HTTP proxy if needed
$ENV{'http_proxy'}= 'http://firewall.net.foi.hr:8001/' ;
#$ENV{'no_proxy'}= '' ;


# Depending on whether you're using Perl 5 or Perl 4 (or want a slight speed
#   advantage), use ONE of the following two blocks.

# Use this for Perl 5 (slows it down a tiny bit).
# You can also remove the "local($1,$2)" line in &fullurl() for slight speedup.
use Socket qw(AF_INET SOCK_STREAM) ;
$AF_INET= AF_INET ; $SOCK_STREAM= SOCK_STREAM ;

# Use this line for Perl 4, or to get a little more speed.  The values of
#   $AF_INET and $SOCK_STREAM must be set correctly for your system; the
#   correct values are usually in either /usr/include/sys/socket.h
#   or /usr/include/linux/socket.h .
#$AF_INET= 2 ; $SOCK_STREAM= 1 ;  # these are GUESSES; set correct values


# Various options to help filter ads and stop cookie-based privacy invasion.
# These are only effective if $FILTER_ADS is set above.
# The default settings are OK, but the sample pattern lists are very simple
#   and could be expanded.  Improvements welcome!
# The champs of ad-filtering are JunkBusters, at http://www.junkbusters.com.
#   If ad-filtering is your primary motive, go there, they da best.
$NO_COOKIE_WITH_IMAGE= 1 ;        # reject cookies attached to images
@BANNED_COOKIE_SERVERS= (         # reject cookies from these servers
    'ad.doubleclick.net', 
    'ad.preferences.com',
    'adforce.imgis.com',
    'ads02.focalink.com',
    ) ;
@BANNED_IMAGE_URL_PATTERNS= (     # reject images matching these URL patterns
    '\b[a-z]\.doubleclick\.net(:\d*)?/',
    '\.imgis.com(:\d*)?/',
    '\.eimg\.com(:\d*)?/',
    '\bexcite\.netscape\.com(:\d*)?/.*/promo/',
    '/excitenetscapepromos/',
    '\bus\.yimg\.com(:\d*)?/(promotions|[a-z])/',
    '\bpromotions\.yahoo\.com(:\d*)?/promotions/',
    '\.cnn\.com(:\d*)?/ads/',
    '/SmartBanner/',
    ) ;
$RETURN_EMPTY_GIF= 1 ;            # replace banned images with 1x1 trans GIF


# Set HTTP_USER_AGENT to something generic like this if you want to be extra
#   careful.  Conceivably, revealing which browser you're using may be a
#   slight privacy or security risk.
# However, note that some URLs serve different pages depending on which
#   browser you're using, so some pages will change if you set this.
#$ENV{'HTTP_USER_AGENT'}= 'Mozilla/4.05 [en] (X11; I; Linux 2.0.34 i586)' ;


#------ end of (useful) user configuration --------------------------------

$PROXY_VERSION= '1.1' ;

# OK, let's time this thing
#($sutime,$sstime)= (times)[0,1] ;

# Fix submitted by Alex Freed:  Under some unidentified conditions,
#   instances of nph-proxy.cgi can hang around for many hours and drag the
#   system.  So until we figure out why that is, here's a 10-minute timeout.
#   Please write me with any insight into this, since I can't reproduce the
#   problem.  Under what conditions, on what systems, does it happen?
# Unfortunately, Windows doesn't seem to support alarms.  Comment out
#   the next two lines if you're using Windows.

$SIG{'ALRM'} = 'timeexit' ;
alarm(600);

# Exit upon timeout.  If you wish, add code to clean up and log an error.
sub timeexit { exit 1 }


# The CGI "standard" at NCSA is unclear on whether $ENV{'SCRIPT_NAME'} 
#   should have a leading slash, and different servers implement
#   it differently.  Here we just make it with a leading slash.
$ENV{'SCRIPT_NAME'}=~ s#^/?#/# ;


# Copy often-used environment vars into scalars, for efficiency
$env_accept= $ENV{'HTTP_ACCEPT'} || '*/*' ;     # may be modified later


# One user reported problems with binary files on certain other OS's, and
#   this seemed to fix it.  Supposedly, either this or the "binmode S" below
#   the newsocketto() call works, or both; I'm putting both in.  Tell me
#   anything new you figure out about this.
binmode STDOUT ;


# %urlsin was an associative array that listed tags and attributes that 
#   may contain URLs, declared like:
# %urlsin= ('a', 'href',
#           'applet', 'codebase',
#           'fig', 'src|imagemap',
#           'form', 'action|script',  ...
#          ) ;


#------ parsing of URL and other input ------------------------------------

# Read the URL from PATH_INFO, stripping the leading slash

$URL= &proxy_decode(substr($ENV{'PATH_INFO'},1)) ;

# Set the query string correctly, from either $ENV{QUERY_STRING} or what's
#   already in $URL.
# The query string may exist in either the encoded URL or in the containing
#   URL, as $ENV{QUERY_STRING}.  If the former, then the query string was
#   (definitely?) in a referenced URL, while the latter most likely implies a
#   GET form input.  Either query string is valid, but form input takes
#   precedence-- if $ENV{QUERY_STRING} exists, it should be used over any
#   query string in the encoded URL.
# Note that Netscape does not pass any query string data that is part of the
#   URL in the <form action> attribute, which is probably correct behaviour.
#   For this program to act the same, it would need to strip the query 
#   string when updating all <form action> URLs, way below.
# Question:  Is there ever a valid case when both QUERY_STRINGs exist??

$URL=~ s/(\?.*)?$/?$ENV{'QUERY_STRING'}/ 
    if $ENV{'QUERY_STRING'} ne '' ;


# Parse the URL, using a regex modelled from the one in RFC 2396 (URI syntax),
#   appendix B.
# Note that $path here will also contain any query component; it's more like
#   the request URI.
# Note that $URL is guaranteed to be an absolute URL with no "#" fragment, 
#   though this does little error-checking.  Note also that the old ";" 
#   parameters are now included in the path component.

($scheme, $host, $port, $path)= 
    ($URL=~ m#^([\w+.-]+)://([^/?:]*):?([^/?]*)(.*)$#i) ;
$host=~ tr/A-Z/a-z/ ;                # for consistency in various places.
$port= 80 if $port eq '' ;
$path= "/$path" if $path!~ m#^/# ;   # if path is '' or contains only query


# Alert the user to non-HTTP URL, with an intermediate page
&nonHTTPwarning($URL) unless ($scheme=~ /^http$/i) ;

# Require a host to be present (for $base_url safety later)
&HTMLdie('The target URL cannot contain an empty host name.') if $host eq '' ;


# Exclude non-text if it's not allowed.  Err on the side of allowing too much.
if ($TEXT_ONLY) {
    # First, forbid requests for filenames with non-text-type extensions
    # Extensions are roughly taken from Netscape's "Helper Preferences" screen
    #   (but that was in 1996).
    # Boy, I really need that list of binary file extensions.  Write me if
    #   you have one (August 1998).
    $nontext= 'gif|jpeg|jpe|jpg|tiff|tif|png|bmp|xbm'   # images
            . '|mp2|mp3|wav|aif|aiff|au|snd'            # audios
            . '|avi|qt|mov|mpeg|mpg|mpe'                # videos
            . '|gz|Z|exe|gtar|tar|zip|sit|hqx|pdf'      # applications
            . '|ram|rm|ra' ;                            # others

    &nontextdie if ($path=~ /\.($nontext)(;|\?|$)/i) ;

    # Then, filter the "Accept:" header to accept only text
    $env_accept=~ s#\*/\*#text/*#g ;    # not strictly perfect
    $env_accept= join(', ', grep(m#^text/#i, split(/\s*,\s*/, $env_accept)) ) ;
    &nontextdie unless $env_accept ne '' ;
}


# Take steps to try removing banner ads
if ($FILTER_ADS) {

    # Determine now if images are banned from this URL (is used later also)
    foreach (@BANNED_IMAGE_URL_PATTERNS) {
	$images_are_banned_here= 1, last if $URL=~ /$_/ ;
    }

    # Intercept request if it looks like an image is requested, i.e. unless
    #   the Accept: header allows either text/... or */... .
    if ($images_are_banned_here) {
	&skip_image unless 
	    grep(m#^(text|\*)/#i, split(/\s*,\s*/, $env_accept) ) ;
    }
}



$realhost= $host ;
$realport= $port ;
$request_uri= $path ;

# there must be a smoother way to handle proxies....
if ($ENV{'http_proxy'}) {
    local($dont_proxy) ;
    foreach (split(/\s*,\s*/, $ENV{'no_proxy'})) {
        $dont_proxy= 1, last if $host=~ /$_$/ ;
    }
    unless ($dont_proxy) {
	# could be slightly more efficient in Perl 5
        ($dummy,$realhost,$realport)=
            $ENV{'http_proxy'}=~ m#^(http://)?([^/?:]*):?([^/?]*)#i ;
        $realport= 80 if $realport eq '' ;
        $request_uri= $URL ;
    }
}


#------ Connect socket to host; send request; wait with select() ----------

# Open a socket and send the request.
# The Accept: and User-Agent: headers are needed because they affect results.
# We're not fully supporting HTTP 1.1, but at least send the Host: header.
# We're anonymously browsing, so don't include From: or Referer: headers.
#   The User-Agent: header is a very slight privacy risk, but some pages
#   load differently with different browsers.

&newsocketto(*S, $realhost, $realport) ;
binmode S ;   # see note with "binmode STDOUT", above

print S $ENV{'REQUEST_METHOD'}, ' ', $request_uri, " HTTP/1.0\015\012",
	'Host: ', $host, ':', $port, "\015\012",    # being a good netizen
        'Accept: ', $env_accept, "\015\012",        # possibly modified above
        'User-Agent: ', $ENV{'HTTP_USER_AGENT'}, "\015\012" ;

# Add the cookie if it exists and cookies are supported.
if ($SUPPORT_COOKIES && ($ENV{'HTTP_COOKIE'} ne '')) {
    $cookie_to_server= &cookie_to_server($ENV{'HTTP_COOKIE'}, $path, $host) ;
    print S 'Cookie: ', $cookie_to_server, "\015\012"
	if $cookie_to_server ne '' ;
}

# If request method is POST, copy content headers and body to request.  Loop
#   to guarantee all is read from STDIN.
if ($ENV{'REQUEST_METHOD'} eq 'POST') {
    $lefttoget= $ENV{'CONTENT_LENGTH'} ;
    print S 'Content-type: ', $ENV{'CONTENT_TYPE'}, "\015\012", 
            'Content-length: ', $lefttoget, "\015\012\015\012" ;
    do {
        $lefttoget-= read(STDIN, $postbody, $lefttoget) ;
        print S $postbody ;
    } while $lefttoget && length($postbody) ;

# For GET requests, just add extra blank line
} else {
    print S "\015\012" ;
}

# Wait a minute for the response to start
vec($rin= '', fileno(S), 1)= 1 ;
select($rin, undef, undef, 60) 
    || &HTMLdie("No response from $realhost:$realport") ;


#------ Read full response into $status, $headers, and $_ (body) ----------

# Full elapsed time, not just CPU time
# $starttime= time ;

# Support both HTTP 1.x and HTTP 0.9
$status= <S> ;  # first line, which is the status line in HTTP 1.x

# HTTP 1.x
if ($status=~ m#^HTTP/#) {
    do {
        $headers.= $_= <S> ;    # $headers includes last blank line
    } until (/^(\015\012|\012)$/) ;   # lines may be terminated with LF or CRLF

    # Unfold long header lines, a la RFC 822 section 3.1.1
    $headers=~ s/(\015\012|\012)[ \t]/ /g ;

    # If we're text only, then cut off non-text responses
    if ($TEXT_ONLY) {
        $*= 1 ;
        if ($headers=~ m#^Content-type:\s*([\w/]*)#i) {
            (close(S), &nontextdie) unless $1=~ m#^text/#i ;
        }
    }

    # If server is one of @BANNED_IMAGE_URL_PATTERNS, then skip the image
    #   unless it's a text type.
    if ($images_are_banned_here) {
	close(S), &skip_image unless ($headers=~ m#^Content-type:\s*text/#i) ;
    }

    $/= '>' ;
    @body= <S> ;

# HTTP 0.9 
} else {
    undef $/ ;
    $body= $status . <S> ;
    $status= '' ;

    # split $body into @body, with each element ending in ">" or end-of-string
    @body= $body=~ /([^>]*>?)/g ;

}

close(S) ;


#------ Modify entire response to point back through this script ----------

# NOTE: IT IS IMPORTANT TO DO THIS AS COMPLETELY AS POSSIBLE.  IF A
# USER UNKNOWINGLY GOES TO PAGE DIRECTLY AND NOT THROUGH THIS PROXY,
# HE/SHE MAY REVEAL HIM/HERSELF IN AN UNINTENDED WAY.

#--------------------------------------------------------------------------
# These were notes to myself from testing the speed of different methods.
#   Names like "nph-proxy2" refer to different modifications.  This version,
#   the fastest, was called nph-proxy2b.
#--------------------------------------------------------------------------
#
# [This version is nph-proxy2b:  Break into @body array, do not use 
#   %urlsin to test if tags should be updated, replacement strings 
#   use (a|b|c) syntax.]
#
# This is by far the most time-consuming part of the script, the updating
#   of all URLs in an HTML file.
#
# Results of informal speed testing (not what I expected):
#   Breaking into @body array instead of one big string definitely saves 
#       significant time-- compare nph-proxy2 to nph-proxy1.  One test
#       showed a time saving of 1/3 to 1/2.
#   Using a %urlsin array does NOT seem to save time, even when only used
#       as boolean test to see if tag might contain URL-- compare nph-proxy2
#       to nph-proxy3, nph-proxy4, and (most similar) nph-proxy5.  Oh well.
#   It seems that reading one tag at a time, converting, and sending it
#       through does NOT save elapsed time over reading all tags before
#       converting, like I thought it would.  Both the CPU and elapsed
#       time are longer for one-tag-at-a-time approach-- compare
#       nph-proxy2 and nph-piper2.
#
#   Results of nph-proxy2 (blocks for each tag) to nph-proxy2b 
#   (single "(att1|att2|att3)" style regex) testing:
#       Mixed results, but overall, using a single regex (nph-proxy2b)
#       takes less CPU "user time", about the same "system time", and
#       slightly more elapsed time than nph-proxy2; I don't know why.
#       The elapsed time is slightly more both within the script and at
#       the HTTP client's end.  All differences are less than 10% on
#       average, and nph-proxy2b occasionally shows LESS elapsed time
#       than nph-proxy2.  CPU time is always less for nph-proxy2b than
#       for nph-proxy2.
#   Since the bottleneck of CPU time is tighter than for elapsed time,
#       and since the elapsed-time loss is less than the CPU time gain,
#       let's go with nph-proxy2b (not that it really makes much difference).
#       Besides, the code is easier to read that way.
#
# So basically, breaking into @body array helps, but not much else does.
#
#--------------------------------------------------------------------------
# 8-4-98 JSM: Found a bug in the regex, so changed it.  It more resembles
#   nph-proxy2.cgi now.
#--------------------------------------------------------------------------

$*= 1 ;     # allow multi-line matching

# Calculate $this_url, $base_host and $base_path here, for use in &fullurl()
#   and &absurl() (including relative URL resolution).
# These are an integral part of those routines, placed here for speed.
#   $this_url is an absolute URL pointing to this proxy script, complete with
#     trailing slash.  It has to be absolute to serve those HTML attributes
#     that require absolute URLs, like <base href>.
#   $base_host has the host/port part of the base URL, with no trailing slash.
#   $base_path has the host/port and path through final slash.

$this_portst=   $ENV{'SERVER_PORT'}==80  ?  ''   :  ':' . $ENV{'SERVER_PORT'} ;
$this_url= join('', 'http://', $ENV{'SERVER_NAME'}, $this_portst, 
                    $ENV{'SCRIPT_NAME'}, '/') ;

# Set $base_url correctly-- first look for a <base> tag, then for a Location: 
#   header; otherwise, use original URL.
foreach (@body) { 
    ($base_url)= m#<\s*base\b[^>]*\bhref\s*=\s*"?([\w+.-]+://[^\s">]+)#i, last
	if /<\s*base\b/i ;
    last if m#<\s*(/head|body)\b#i ;     # stop looking after </head> or <body>
}
($base_url)= $headers=~ m#^Content-Base:\s*([\w+.-]+://\S+)#i if $base_url eq '' ;
($base_url)= $headers=~ m#^Content-Location:\s*([\w+.-]+://\S+)#i if $base_url eq '' ;
($base_url)= $headers=~ m#^Location:\s*([\w+.-]+://\S+)#i if $base_url eq '' ;
$base_url  = $URL                                         if $base_url eq '' ;

# Guarantee that $base_url has at least a path of '/', inserting before ?query
#   if needed.
$base_url=~ s#^([\w+.-]+://[^/?]+)/?#$1/# ;

($base_host)= $base_url=~ m#^([\w+.-]+://[^/?]+)# ;   # does not end in slash
($base_path)= $base_url=~ m#^([^?]*/)# ;              # rely on greedy matching


# First, modify any appropriate HTTP headers to point back through the proxy.
#   Location: is most important.
$headers=~ s/^Content-Base:\s*(\S*)/    'Content-Base: '    .&fullurl($1)/gie ;
$headers=~ s/^Content-Location:\s*(\S*)/'Content-Location: '.&fullurl($1)/gie ;
$headers=~ s/^Location:\s*(\S*)/        'Location: '        .&fullurl($1)/gie ;
$headers=~ s/^URI:\s*(\S*)/             'URI: '             .&fullurl($1)/gie ;

# And the non-standard Refresh: header... any others?
$headers=~ s/^(Refresh:.*;\s*URL=)(\S*)/ $1                 .&fullurl($2)/gie ;


# Modify cookies to point back through the script, or they won't work.
# If they're not supported, then filter them all out.
if ($SUPPORT_COOKIES) {

    # Take appropriate action if we're trying to filter ads out
    if ($FILTER_ADS) {

	# If so configured, filter out cookies packaged with non-text resources
	# This actually covers all non-text resources, not just images.
	if ($NO_COOKIE_WITH_IMAGE) {
	    $headers=~ s/^Set-Cookie:.*\012//gi
		unless $headers=~ m#^Content-Type:\s*text/#i ;
	}

	# Strip the cookies if they come from one of @BANNED_COOKIE_SERVERS
	foreach (@BANNED_COOKIE_SERVERS) {
	    $headers=~ s/^Set-Cookie:.*\012//gi, last    if $host eq $_ ;
	}
    }

    $headers=~ s/^Set-Cookie:\s*([^\015\012]*)/
	         'Set-Cookie: ' . &cookie_to_client($1, $path, $host)
		/gie ;

} else {
    $headers=~ s/^Set-Cookie:.*\012//gi ;
}



# Update all URLs in all tags that refer to URLs
# Only update the URLs if it's HTML (or using HTTP 0.9), and if it's not
#   empty.

if ( (($headers=~ m#^Content-type:\s*text/html\b#i) || ($headers eq '') )
     && ($body[0] ne '') ) {

    # This is a complete list of HTML tags/attributes that may include a URL,
    #   compiled from the list at Sandia (read on 9-17-96, but last updated on
    #   12-7-95), and the latest HTML DTD as of 9-17-96.  The Sandia list
    #   includes non-standard tags found to be used by Netscape or Microsoft.
    # List updated 7-31-98 by JSM to include all URL-type attributes defined
    #   in HTML 4.0.
    # If anyone knows of a well-maintained list of standard and non-standard
    #   tags/attributes with URLs in them, please let me know!!
    # Note that most of these are very rarely used, if ever.  They're included
    #   for safety, since we don't want an anonymous user accidentally
    #   revealing themselves because of a non-anonymized URL.
    # This regex favors speed over readability.
    # This could probably be more efficient, but the solution eludes me.
    # Note that this is not a perfect regex.  The substitutions can be fooled
    #   by, e.g., <input value="<a href=bad"> or <input value="src=bad">.
    #   Also, it doesn't handle single-quoted attributes, which I have
    #   rarely seen.  It could be fixed with a slower regex.
    # The regex also terminates URLs at spaces (a bug); to fix this would 
    #   take a much slower and messier regex.
    # Denoting which of these are for images/binaries might be helpful, if
    #   we need more elaborate text-only support.
    # 8-4-98 JSM:  Changed the regex to fix a bug: would only change the 
    #   first URL in a tag.  New regex is probably slower.

$nuke_body=0;

    foreach (@body) {

	
	# nuke deja.com shit

#		print "-b($nuke_body)-$_-e-\n";

		if (/<!-- begin CONTENT here -->/i) {
			$nuke_body=0;
		}
		if (/<!-- second stage content : begin -->/i) {
			$nuke_body=0;
		}

		if ($nuke_body) {
			s/.*//gi;
			next;
		}

		if (/<!-- PLOC/i) {
			$nuke_body++;
		}
		if (/<!-- end CONTENT here -->/) {
			$nuke_body++;
		}
		if (/<!-- second stage content : end -->/) {
			$nuke_body++;
		}

	# fix max hist to 100 :-)
	s/<option value="25" selected>/<option value="25">/;
	s/<option value="100">/<option value="100" selected>/;

	# Put the most common cases first

        s/(<[^>]*\bhref\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*a\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\blowsrc\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
        s/(<[^>]*\blongdesc\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
        s/(<[^>]*\busemap\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
        s/(<[^>]*\bdynsrc\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
            next if /<\s*img\b/i ;

        s/(<[^>]*\bbackground\s*=\s*"?)([^\s">]*)/ $1 . &fullurl($2) /ie,
            next if /<\s*body\b/i ;

        s/(<[^>]*\bhref\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*base\b/i ;     # has special significance

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\blongdesc\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
            next if /<\s*frame\b/i ;

        s/(<[^>]*\baction\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
        s/(<[^>]*\bscript\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
            next if /<\s*form\b/i ;     # needs special attention

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\busemap\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
            next if /<\s*input\b/i ;

        s/(<[^>]*\bhref\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*area\b/i ;

        s/(<[^>]*\bcodebase\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
        s/(<[^>]*\bcode\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
        s/(<[^>]*\bobject\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
        s/(<[^>]*\barchive\s*=\s*"?)([^\s">]*)/    $1 . &fullurl($2) /ie,
            next if /<\s*applet\b/i ;


        # These are seldom-used tags, or tags that seldom have URLs in them

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*bgsound\b/i ;  # Microsoft only

        s/(<[^>]*\bcite\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*blockquote\b/i ;

        s/(<[^>]*\bcite\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*del\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*embed\b/i ;    # Netscape only

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\bimagemap\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
            next if /<\s*fig\b/i ;      # HTML 3.0

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*h[1-6]\b/i ;   # HTML 3.0

        s/(<[^>]*\bprofile\s*=\s*"?)([^\s">]*)/    $1 . &fullurl($2) /ie,
            next if /<\s*head\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*hr\b/i ;       # HTML 3.0

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\blongdesc\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
            next if /<\s*iframe\b/i ;

        s/(<[^>]*\bcite\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*ins\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*layer\b/i ;

        s/(<[^>]*\bhref\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
        s/(<[^>]*\burn\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*link\b/i ;

        s/(<[^>]*\burl\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*meta\b/i ;     # Netscape only

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*note\b/i ;     # HTML 3.0

        s/(<[^>]*\busemap\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
        s/(<[^>]*\bcodebase\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
        s/(<[^>]*\bdata\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
        s/(<[^>]*\barchive\s*=\s*"?)([^\s">]*)/    $1 . &fullurl($2) /ie,
        s/(<[^>]*\bclassid\s*=\s*"?)([^\s">]*)/    $1 . &fullurl($2) /ie,
        s/(<[^>]*\bname\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*object\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\bimagemap\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
            next if /<\s*overlay\b/i ;  # HTML 3.0

        s/(<[^>]*\bcite\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*q\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\bfor\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*script\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*select\b/i ;   # HTML 3.0

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*ul\b/i ;       # HTML 3.0


    }   # foreach (@body)


    # Does this ever cause a problem, putting a comment before the entire body?
    unshift(@body, "<!-- resource has been modified by proxy script -->\n" ) ;

    # Change Content-Length header, since we changed the content
    $headers=~ s/^Content-Length:.*\012/ 
                 'Content-Length: ' . length(join('',@body)) . "\015\012"/ie ;

}

# print the status line, headers, and the entire (modified) resource
print $status, $headers, @body ;


# Put this back in to run speed trials
#if (($headers=~ m#^Content-type:\s*text/html#i) || !$headers) {
#    # OK, let's time this thing
#    ($eutime,$estime)= (times)[0,1] ;
#    open(LOG,">>proxy2b.log") ;
#    print LOG "full times: ", $eutime-$sutime, " ", $estime-$sstime, 
#        " ", time-$starttime, "  URL: $URL\n" ;
#    close(LOG) ;
#}


exit ;



# Returns the full absolute URL to query our script for the given URI
#   reference.  PATH_INFO will include the encoded absolute URL of the target,
#   but the fragment will be appended unencoded so browsers will resolve it
#   correctly.
# Note that the calculations of $this_url, $base_host, and $base_path above
#   are an integral part of this routine, placed above for speed.
# This was rewritten 2-18-99 to be much more flexible, e.g. to allow encoding
#   and avoid illegal PATH_INFO sequences.  Unfortunately, it is a little 
#   slower.
# The URL to be encoded should include everything that is sent in the HTTP
#   request, including any query, but not any fragment.
# This only returns absolute URLs, though relative URLs would usually suffice.
#   If it matters, we could have a fullrelurl() and fullabsurl(), the latter
#   used for those HTML attributes that require an absolute URL (like <base>).
#
# The ?:?: statement resolves relative URLs to absolute URLs, given the
#   $base_{url,host,path} variables figured earlier.  It does it simply and
#   efficiently, and accurately enough; the full procedure is described in 
#   RFC 2396 (URI syntax), section 5.2.
# RFC 2396, section 5 states that there are three types of relative URIs:
#   net_path (beginning with //, rarely used), abs_path (beginning with /),
#   and rel_path, any of which may be followed by a "?query"; the query must
#   be included in the result.  Thus, we only need to examine the start of
#   the relative URL.  (Only the latter two types are supported below.)
# This ?:?: statement passes all test cases in RFC 2396 appendix C, except
#   for the following:  It does not reduce . and .. path segments (to do
#   so would take a lot more time); it ignores the rare net_path possibility
#   for speed, though that could be added easily with an extra case here and
#   a $base_scheme calculation above; and it assumes $uri_ref has something
#   other than an empty fragment in it, i.e. that the URI is non-empty.

sub fullurl {
    local($uri_ref)= @_ ;
    local($1,$2) ;   # Required for Perl 4-- else, caller's $1,$2 get clobbered
    local($uri,$frag)= $uri_ref=~ /^([^#]*)(#.*)?/ ;
    return $uri_ref if $uri eq '' ;  # allow bare fragments to pass unchanged

    # calculate absolute URL based on three possible cases
    local($absurl)= 
            $uri=~ m#^[\w+.-]*:#i   ?  $uri                # absolute URL
          : $uri=~ m#^/#            ?  $base_host . $uri   # abs path, rel URL
          :                            $base_path . $uri ; # relative path

    return join('', $this_url, &proxy_encode($absurl), $frag ) ;
}


#------ Cookie routines ---------------------------------------------------

# Cookie support:  The trick is how to send a cookie back to the client that
#   it will return for appropriate hosts.  Given that the target URL may be
#   encoded, and the client can't always tell where the target URL is, the 
#   only way to do that is to get *all* the cookies from the client and send
#   along the matching ones.  If the client has a lot of cookies through the
#   proxy, this could conceivably be a problem.  Oh well, it works for the
#   limited amount I've tested.
# Here, we transform the cookie from the server into something the client
#   will always send back to us, and embed the real server/path info in the
#   name of the name-value pair, since the cookie is uniquely identified by
#   the domain, path, and name.  Upon return from the client, we split the
#   name back into its original fields.
# One way to get around *some* of the all-cookies-all-the-time problem,
#   *sometimes*, may be possible to program with the following approach:
#   First, the target URL must be "encoded" (in proxy_encode()) in a way
#   that it resembles a path.  For example, the default "://" --> "/"
#   encoding does this.  Then, let the cookies go back to the client with
#   the target paths still intact.  This would only work when the cookie
#   domain is the default, i.e. the source host.  Check other possibilities
#   carefully, too, but I think you could get it to work somehow.
# Question-- is the port supposed to be used in the domain field?
#   Everything here assumes not, which is conceivably a security risk.

# Transform the cookie into something the client will send back through
#   the script, but still has all the needed info.  Returns undef if
#   cookie is invalid (e.g. comes from the wrong host).
# A cookie is uniquely identified by the domain, path, and name, so this 
#   transformation embeds the path and domain info into the "name".
sub cookie_to_client {
    local($cookie, $source_path, $source_server)= @_ ;
    local($name, $value, $expires_clause, $path, $domain, $secure_clause) ;
    local($new_name, $new_cookie) ;

    ($name,$value)= $cookie=~ /^\s*([^=;,\s]*)=?([^;,\s]*)/ ;
    ($expires_clause)= $cookie=~ /;\s*(expires\s*=[^;]*)/i ;
    ($path)= $cookie=~ /;\s*path\s*=\s*([^;,\s]*)/i ;      # clash w/ ;-params?
    ($domain)= $cookie=~ /;\s*domain\s*=\s*([^;,\s]*)/i ;
    ($secure_clause)= $cookie=~ /;(\s*secure\b)/i ;

    # Path defaults to path of URL that sent the cookie
    $path= $source_path if $path eq '' ;

    # Domain must be checked for validity: defaults to the server that sent
    #   the cookie; otherwise, must match end of that server name, and must
    #   contain at least two dots if in one of these seven top-level domains, 
    #   three dots otherwise.
    if ($domain eq '') {
	$domain= $source_server ;
    } else {
	return(undef) if $source_server!~ /$domain$/ ;
	return(undef) unless 
	    ( ( ($domain=~ tr/././) >= 3 ) ||
	      ( ($domain=~ tr/././) >= 2 && 
	        $domain=~ /\.(com|edu|net|org|gov|mil|int)$/ )
	    ) ;
    }

    # This is hereby the transformed format: name is $name;$path;$domain
    #   with %, ;, and = URL-encoded.  (The three values won't already
    #   have semicolons in them.)
    $new_name= join(';', $name, $path, $domain) ;
    $new_name=~ s/%/%25/g ;    # must be first
    $new_name=~ s/;/%3b/g ;
    $new_name=~ s/=/%3d/g ;

    # The new domain is this proxy server, which is the default if it is not 
    #   specified.
    $new_cookie= join('; ', grep(length,
				 $new_name . '=' . $value, 
				 $expires_clause, 
				 'path=' . $ENV{'SCRIPT_NAME'} . '/',
				 $secure_clause 
		     )) ;

    return($new_cookie) ;

}


# Unwrap the cookies from the client that were transformed earlier by the
#   routine cookie_to_client().  Returns composite cookie with all cookies
#   that match the domain and path, and no others!
# Currently, the name field contains the path and domain info-- see
#   cookie_to_client() for description of format.
sub cookie_to_server {
    local($cookie, $target_path, $target_server)= @_ ;
    local(@cookies)= split(/\s*;\s*/, $cookie) ;
    local($name, $value, $path, $domain, @matches, %pathlen) ;
    foreach (@cookies) {
	($name,$value)= split(/=/, $_, 2) ;
        $name=~ s/%3d/=/g ;
	$name=~ s/%3b/;/g ;
	$name=~ s/%25/%/g ;
	($name,$path,$domain)= split(/;/, $name) ;
	push(@matches, $name.'='.$value), 
	    $pathlen{$matches[$#matches]}= length($path)
		if ($target_server=~ /$domain$/i  && $target_path=~ /^$path/) ;
    }

    # More specific path mappings (i.e. longer paths) should be sent first.
    sub reversepathlength { length($pathlen{$b}) <=> length($pathlen{$a}) }
    return join('; ', sort reversepathlength @matches) ;
}



#------ Utility and output routines ---------------------------------------

# The following subroutine looks messy, but can be used to open any
#   TCP/IP socket in any Perl program.  Except for the &HTMLdie() part.
# Note that $AF_INET and $SOCK_STREAM are set near the start of the program,
#   near the "use Socket" statement.
sub newsocketto {
    local(*S, $host, $port)= @_ ;
    local($hostaddr,$remotehost) ;

    # Create the remote host data structure, from host name or IP address
    $hostaddr= ($host=~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/)
                  ?  pack('c4', $1, $2, $3, $4)     # for IP address
                  :  ( (gethostbyname($host))[4]    # for alpha host name
                       || &HTMLdie("Couldn't find address for $host: $!") ) ;
    $remotehost= pack('S n a4 x8', $AF_INET, $port, $hostaddr) ;

    # Create the socket and connect to the remote host
    socket(S, $AF_INET, $SOCK_STREAM, (getprotobyname('tcp'))[2])
        || &HTMLdie("Couldn't create socket: $!") ;
    connect(S, $remotehost) 
        || &HTMLdie("Couldn't connect to $host:$port: $!") ;
    select((select(S), $|=1)[0]) ;      # unbuffer the socket
}


# Alert the user to non-HTTP URL, with this intermediate page
sub nonHTTPwarning {
    print <<EOF ;
HTTP/1.0 200 OK
Content-type: text/html

<html>
<head><title>WARNING: Entering non-anonymous area!</title></head>
<body>
<h1>WARNING: Entering non-anonymous area!</h1>
<h3>This proxy only supports HTTP.  Any browsing to a non-HTTP URL will
be directly from your browser, and no longer anonymous.</h3>
<h3>Click the link below to continue to the URL, non-anonymously.</h3>
<blockquote><tt><a href="$_[0]">$_[0]</a></tt></blockquote>
<p>
<hr>
<a href="http://www.jmarshall.com/tools/cgiproxy/"><i>CGIProxy $PROXY_VERSION</i></a>
<p>
</body>
</html>
EOF

    exit ;
}


# Return "403 Forbidden" message, with explanatory text
sub nontextdie {
    print <<EOF ;
HTTP/1.0 403 Forbidden
Content-type: text/html

<html>
<head><title>Proxy cannot forward non-text files</title></head>
<body>
<h1>Proxy cannot forward non-text files</h1>
<p>Due to bandwidth limitations, the owner of this particular proxy is
forwarding only text files.  For best results, turn off automatic image 
loading if your browser lets you.
<p>If you need access to images or other binary data, route your browser 
through another proxy (or install one yourself--
<a href="http://www.jmarshall.com/tools/cgiproxy/">it's easy</a>).
<p>
<hr>
<a href="http://www.jmarshall.com/tools/cgiproxy/"><i>CGIProxy $PROXY_VERSION</i></a>
<p>
</body>
</html>
EOF

    exit ;
}


# When an image should be blanked, returns either a transparent 1x1 GIF or
#   a 404 result.  Yes, that's an inlined 43-byte GIF.
sub skip_image {
    if ($RETURN_EMPTY_GIF) {
	print <<EOF ;
HTTP/1.0 200 OK
Content-Type: image/gif
Content-Length: 43

GIF89a\x01\0\x01\0\x80\0\0\0\0\0\xff\xff\xff\x21\xf9\x04\x01\0\0\0\0\x2c\0\0\0\0\x01\0\x01\0\x40\x02\x02\x44\x01\0\x3b
EOF
    } else {
	print "HTTP/1.0 404 Not Found\n\n" ;
    }

    exit ;
}


# Die, outputting HTML error page
sub HTMLdie {
    local($msg)= @_ ;
    print <<EOF ;
HTTP/1.0 200 OK
Content-Type: text/html

<html>
<head><title>CGI Proxy Error</title></head>
<body>
<h1>CGI Proxy Error</h1>
<h3>$msg</h3>
<p>
<hr>
<a href="http://www.jmarshall.com/tools/cgiproxy/"><i>CGIProxy $PROXY_VERSION</i></a>
<p>
</body>
</html>
EOF

    exit ;
}
